import glob
import os
import json
import yaml
import re
import collections
import json
import curses
import shlex
import tabulate
import itertools
import time
from contextlib import contextmanager
from pathlib import Path

# GPU sorting
def query_gpus():
    """
    return dictionary structure:
    {'0': {'gpu_name'       : 'GeForce RTX 2080 Ti',
           'memory.free'    : 11017,
           'memory.used'    : 1,
           'memory.total'   : 11018,
           'utilization.gpu': 0
           },
     '1': {'gpu_name'       : 'GeForce RTX 2080 Ti',
           'memory.free'    : 11018,
           'memory.used'    : 1,
           'memory.total'   : 11019,
           'utilization.gpu': 0
           }
     }
    """
    query_args = ['index', 'gpu_name', 'memory.free', 'memory.used', 'memory.total', 'utilization.gpu']
    cmd = 'nvidia-smi --query-gpu={} --format=csv,noheader'.format(','.join(query_args))

    def parse(key, value):
        if key in ['memory.free', 'memory.total', 'memory.used']:
            return int(value.upper().strip().replace('MIB', ''))
        elif key == 'utilization.gpu':
            return int(value.replace('%', '').strip())
        else:
            return value.strip()

    gpu_infos = {}
    for info in os.popen(cmd).readlines():
        info_dict = {key: parse(key, value)
                     for key, value in zip(query_args, info.strip().split(','))}
        index = info_dict['index']
        del (info_dict['index'])
        gpu_infos[index] = info_dict
    return gpu_infos


def sort_gpus(gpus, min_mem=None):
    """
    Sort a list of single gpus
    """
    gpus = list(set(gpus))
    gpu_infos = query_gpus()
    if min_mem is not None:
        valid_gpus = [(gpu_infos[gpu]["memory.free"], gpu) for gpu in gpus if
                      gpu_infos[gpu]["memory.free"] > min_mem]
        if not valid_gpus:
            raise ValueError("No GPU with expected memory {} MB.".format(min_mem))

    # fuzzy cmp
    if min_mem is not None:
        gpus.sort(key=lambda x: (not gpu_infos[x]["memory.free"] > min_mem, gpu_infos[x]['utilization.gpu'] // 4))
    else:
        gpus.sort(key=lambda x: (gpu_infos[x]["memory.used"] // 100, gpu_infos[x]['utilization.gpu'] // 4))
    return gpus


# naming
def snake2camel(snake_str, shrink_keep=0):
    """
    "a_snake_case_string" or "a-snake-case-string" to "ASnakeCaseString"
    if shrink_keep > 0, say shrink_keep = 2
    "a_snake_case_string" to "ASnCaString"
    """
    components = snake_str.split('-')
    if len(components) == 1:
        components = components[0].split('_')
    if shrink_keep:
        return ''.join([x[0:shrink_keep].title() if len(x) > shrink_keep else x
                        for x in components[:-1]]) + components[-1].title()
    else:
        return ''.join(x.title() for x in components)


def entry2str(name, value, str_maxlen, basedir=True):
    name = snake2camel(name, shrink_keep=2)
    if isinstance(value, str):
        if os.path.exists(value) and basedir:
            value = os.path.basename(value)
        else:
            # avoid directory split when used as directory name
            value = re.sub("^[/.]*", "", value)  # avoid leading "." and "/"
            value = re.sub("/", "_", value)
        if len(value) > str_maxlen:
            value = value[-str_maxlen:]
    elif isinstance(value, bool):
        if value:
            value = "T"
        else:
            value = "F"
    elif isinstance(value, (float, int)):
        value = "%g" % value
    else:
        value = str(value)
    return name + '_' + value


def spec2name(spec, str_maxlen, basedir=True):
    keys = list(spec.keys())
    keys.sort()
    strs = [entry2str(key, value, str_maxlen, basedir) for key, value in spec.items()]
    name = '-'.join(strs)
    return name


def shell_arg(value):
    if isinstance(value, str):
        return shlex.quote(value)
    elif isinstance(value, (int, float)):
        return str(value)
    else:
        raise ValueError("{} is not a str, int or float.".format(repr(value)))


def map_placeholder(spec, param):
    value = spec[param]
    if isinstance(value, (str, int, float)):
        return shell_arg(value)
    elif isinstance(value, bool):
        ValueError(
                "boolean param {} no supported as {}, please specify it as {}.".format(repr(param),
                                                                                       repr("{" + param + "}"),
                                                                                       repr("[" + param + "]")))
    else:
        raise ValueError("{} of type {} is supported. use str, int, or float.".format(repr(param), type(value)))


def map_replacement(spec, param):
    value = spec[param]
    if isinstance(value, bool):
        if value:
            return '--{}'.format(param)
        else:
            return ""
    elif isinstance(value, (str, int, float)):
        return "--{} ".format(param) + shell_arg(value)
    else:
        raise ValueError("{} of type {} is supported. use str, bool, int, or float.".format(repr(param), type(value)))


def json_load(filename):
    with open(filename, "r") as fin:
        d = json.load(fin)
    return d


def json_dump(d, filename):
    with open(filename, "w") as fout:
        json.dump(d, fout, sort_keys=True, indent=4)


@contextmanager
def lock(filename, poll=0.2, timeout=60):
    while timeout > 0:
        try:
            os.link(filename, filename + ".lock")
            break
        except FileExistsError:
            time.sleep(poll)
            timeout -= poll

    try:
        yield filename

    finally:
        try:
            os.unlink(filename + ".lock")
        except FileNotFoundError:
            pass


def yaml_load(filename):
    with open(filename, "r") as fin:
        d = yaml.load(fin, Loader=yaml.FullLoader)
    return d


def yaml_dump(d, filename):
    with open(filename, "w") as fout:
        yaml.dump(d, stream=fout, width=1000000)


@contextmanager
def edit_yaml(output, file):
    # read&write yaml with lock
    path = Path(output, file)
    if not path.exists():
        path.touch()
    yaml_dict = {}
    with lock(str(path)):
        try:
            yaml_dict = yaml_load(path)
            if yaml_dict is None:
                yaml_dict = {}
            # pass by reference. re-assigned object in the `with` context will not be saved.
            yield yaml_dict
        finally:
            if yaml_dict is not None:
                yaml_dump(yaml_dict, path)


def color_print(str, color):
    color_map = {
        "RED":     "\x1b[31m",
        "GREEN":   "\x1b[32m",
        "YELLOW":  "\x1b[33m",
        "BLUE":    "\x1b[34m",
        "MAGENTA": "\x1b[35m",
        "CYAN":    "\x1b[36m",
        "WHITE":   "\x1b[37m"
    }
    print(color_map[color.upper()] + str + "\x1b[0m")



import collections
import glob
import re
import os
import tabulate
from ..utils.misc import yaml_load

Experiment = collections.namedtuple("Experiment", ["cache", "metric", "param"])


def trunc(x, significant=2):
    return float(format(x, '.{}f'.format(significant)))


def dict2tuple(dictionary, keys):
    return tuple(dictionary.get(k, None) for k in keys)


def list2tsv(t):
    return "\t".join(str(i) for i in t)


def table(entries, headers, tsv=True):
    if tsv:
        print(list2tsv(headers))
        for entry in entries:
            if isinstance(entry, dict):
                print(list2tsv(dict2tuple(entry, headers)))
            elif isinstance(entry, (list, tuple)):
                print(list2tsv(entry))
    else:
        print(tabulate.tabulate(entries, headers, "pipe"))


def mean_std_num(values, significant=2, scale=1):
    num = len(values)
    if num == 0:
        return 0, 0, 0
    mean = sum(values) / num
    std = (sum((v - mean) ** 2 for v in values) / num) ** 0.5
    return trunc(mean * scale, significant), trunc(std * scale, significant), num


class Key2Values(object):
    # Group values with the same key
    def __init__(self, group_func=list):
        assert group_func in [list, set]
        self.group_func = group_func
        self.group = collections.defaultdict(group_func)

        # redirect methods
        self.items = self.group.items
        self.keys = self.group.keys
        self.values = self.group.values
        self.get = self.group.get
        self.__contains__ = self.group.__contains__
        self.__delitem__ = self.group.__delitem__
        self.__sizeof__ = self.group.__sizeof__

    def __getitem__(self, item):
        return self.group[item]

    def __setitem__(self, key, value):
        self.group[key] = value

    def __repr__(self):
        return repr(self.group).replace("defaultdict", "Key2Values")

    def __len__(self):
        return len(self.group)

    def add(self, k, v):
        if self.group_func == set:
            if isinstance(v, (list, tuple)):
                self.group[k].update(v)
            else:
                self.group[k].add(v)
        else:
            if isinstance(v, (list, tuple)):
                self.group[k].extend(v)
            else:
                self.group[k].append(v)


class Examiner(object):
    """
    Example usage:

    def func(path, experiment, caches):
        log = latest_log("test", path)
        if log is None:
            return
        experiment.cache["cache1"] = get_cache1(log)
        experiment.metric["metric1"] = get_metric1(log)
        caches["global_cache1"] = get_global_cache1(log)

    examiner = Examiner()
    examiner.add(func)
    examiner.exam("output", "*")
    examiner.table()

    """

    def __init__(self):
        self.exams = set()
        self.caches = {}
        self.experiments = {}

    def reset(self):
        self.exams = set()
        self.caches = {}
        self.experiments = {}

    def add(self, func):
        self.exams.add(func)

    def exam(self, output="output", regex=".*", verbose=False):
        for i, path in enumerate(match_output(output, regex)):
            params = load_params(path)
            if params is None:
                print("No 'params.yaml' found, skip {}".format(path))
                continue
            exp = os.path.basename(path)
            if verbose:
                print("{}: {}".format(i, exp))
            if exp in self.experiments:
                experiment = self.experiments[exp]
            else:
                experiment = Experiment(cache={}, metric={}, param={})
            experiment.param.update(params)
            for exam in self.exams:
                exam(path, experiment, self.caches)
            self.experiments[exp] = experiment

    def _get_param2values(self):
        param2values = Key2Values(set)
        for experiment in self.experiments.values():
            for k, v in experiment.param.items():
                if re.match("_", k):
                    continue
                param2values.add(k, v)
        return param2values

    def table(self, tsv=True, concise=True, raw=False):
        # find unique param values
        param2values = self._get_param2values()
        if concise:
            # remove columns with the same values
            param_headers = sorted(list(k for k, v in param2values.items() if len(v) > 1))
        else:
            param_headers = sorted(list(param2values.keys()))

        metric_headers = set()
        for experiment in self.experiments.values():
            metric_headers.update(experiment.metric.keys())
        metric_headers = sorted(list(metric_headers))
        headers = param_headers + metric_headers
        entries = [[experiment.param.get(h, None) for h in param_headers] +
                   [experiment.metric.get(h, None) for h in metric_headers]
                   for experiment in self.experiments.values()]
        table(entries, headers, tsv)
        if raw:
            return entries, headers

    def table_agg(self, metric, params, scale=1, significant=2, tsv=True, concise=True, raw=False):
        # find unique param values
        param2values = self._get_param2values()
        if concise:
            # remove columns with the same values
            param_headers = sorted(list(k for k, v in param2values.items() if len(v) > 1))
        else:
            param_headers = sorted(list(param2values.keys()))

        if not isinstance(params, (list, tuple)):
            params = [params]
        param_headers = [h for h in param_headers if h not in params]

        k2v = Key2Values(list)
        for experiment in self.experiments.values():
            k = tuple(experiment.param.get(h, None) for h in param_headers)
            v = experiment.metric.get(metric, None)
            if v is None:
                continue
            else:
                k2v.add(k, v)
        headers = param_headers + ["mean", "std", "num"]
        entries = []
        for k, v in k2v.items():
            entries.append((*k, *mean_std_num(v, significant, scale)))
        table(entries, headers, tsv)
        if raw:
            return entries, headers


def match_output(output="output", regex=".*"):
    pattern = re.compile(regex)
    files = [file for file in os.listdir(output) if pattern.search(file)]
    paths = [os.path.join(output, file) for file in files]
    paths = [path for path in paths if os.path.isdir(path)]
    return sorted(paths)


def load_params(path):
    param_path = os.path.join(path, "param")
    if not os.path.exists(param_path):
        return None
    params = yaml_load(param_path)
    return params


def latest_log(command, path, index=-1):
    paths = sorted(glob.glob(os.path.join(path, "log.{}.*".format(command))))
    num_logs = len(paths)
    assert isinstance(index, int)
    if index < -num_logs or index > num_logs - 1:
        return None
    return paths[index]



# -*- coding: utf-8 -*-
import asyncio
import argparse
import random
import itertools
import shlex
import re
import types
import copy
import sys
import os
import shutil
from datetime import datetime
from runner.utils.misc import spec2name, map_placeholder, map_replacement, shell_arg, \
    yaml_load, yaml_dump, color_print, edit_yaml
from runner.utils.config import load_yaml, InvalidYAMLException

TIME = datetime.now().strftime("%Y%m%d.%H%M%S")


def run(coro):
    if sys.version_info >= (3, 7):
        return asyncio.run(coro)
    # Emulate asyncio.run() on older versions
    # asyncio.run() requires a coroutine, so require it here as well
    if not isinstance(coro, types.CoroutineType):
        raise TypeError("run() requires a coroutine object")

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        return loop.run_until_complete(coro)
    finally:
        loop.close()
        asyncio.set_event_loop(None)


def sweep(param2choices, num_sample=None):
    # random combination of param choices
    for key, values in param2choices.items():
        if not isinstance(values, list):
            raise InvalidYAMLException("{} should be a list, not {}".format(key, type(values)))

    param_choices = [[(key, value) for value in values] for key, values in param2choices.items()]
    cand_params = [list(param) for param in itertools.product(*param_choices)]
    if num_sample is not None:
        random.shuffle(cand_params)
        params = cand_params[:num_sample]
        params.sort()
    else:
        params = cand_params
    param_dicts = [dict(param) for param in params]
    return param_dicts


# param_dict updates
def map_alias(param_dict, aliases):
    def resolve_alias(key, value):
        if key not in aliases:
            raise InvalidYAMLException("{} not in 'alias'.")
        # aliases[key] already asserted as a dict
        if value not in aliases[key]:
            raise InvalidYAMLException("value of '{}' should be one of {}".format(key, tuple(aliases[key].keys())))
        return aliases[key][value]

    update_dict = dict()
    alias_param = set()
    for param, value in param_dict.items():
        if param in aliases:
            alias_param.add(param)
            update_dict.update(resolve_alias(param, value))
    for param in alias_param:
        del param_dict[param]
    param_dict.update(update_dict)


def add_default(param_dict, defaults):
    for param, value in defaults.items():
        if param not in param_dict:
            param_dict[param] = value


def parse_choice(args, choice, aliases, defaults):
    # update commands
    if "_cmd" in choice:
        commands = choice["_cmd"]
        del choice["_cmd"]
    else:
        commands = None
    entries = []
    for spec in sweep(choice, num_sample=args.sample):
        name = spec2name(spec, str_maxlen=100)
        meta = {
            "_name":   name,
            "_time":   TIME,
            "_output": args.output if args.no_subdir else os.path.join(args.output, name)
        }
        map_alias(spec, aliases)
        add_default(spec, defaults)
        entries.append((spec, meta))
    return commands, entries


def build_tasks(args, templates, aliases, defaults, choices):
    def uniq_entries(entries):
        for spec, meta in entries:
            if spec in uniq_entries.uniq_spec:
                # allow the same task with different command.
                # repeated commands will be skipped later by checking stat
                if meta["_output"] not in uniq_entries.uniq_output:
                    continue
            else:
                uniq_entries.uniq_spec.append(copy.deepcopy(spec))
                uniq_entries.uniq_output.append(meta["_output"])
            yield spec, meta

    uniq_entries.uniq_spec = []
    uniq_entries.uniq_output = []

    tasks = []
    orphans = set()  # track params not consumed by any command
    for choice in choices:
        commands, entries = parse_choice(args, choice, aliases, defaults)
        if args.command:
            # command line option will override those in the yaml config
            commands = args.command
        entries = uniq_entries(entries)
        for spec, meta in entries:
            unused_params = set(spec.keys())  # avoid accidentally missing the param in the template
            spec.update(meta)
            scripts = {}
            for command, template in templates.items():
                if commands and command not in commands:
                    continue
                template = re.sub(r"\s*\n\s*", " ", template)  # clean up line breaks
                # fill placeholder with params
                empties = set()
                placeholders = re.findall(r"{[\w\-\_]+?}", template)  # {param} -> value
                replacements = re.findall(r"\[[\w\-\_]+?\]", template)  # [param] -> --param value
                intersect = set(p.strip("{}") for p in placeholders) & set(r.strip("[]") for r in replacements)
                if intersect:
                    raise InvalidYAMLException("duplicate params in placeholder and replacement: {}".format(intersect))

                for placeholder in placeholders:
                    param = placeholder.strip("{}")
                    if param in aliases:
                        raise InvalidYAMLException(
                                "alias param '{}' should not be specified in template.".format(param))
                    if param in unused_params:
                        unused_params.remove(param)
                    if param in spec:
                        template = template.replace(placeholder, map_placeholder(spec, param))
                    else:
                        empties.add(placeholder)

                for replacement in replacements:
                    param = replacement.strip("[]")
                    if param in aliases:
                        raise InvalidYAMLException(
                                "alias param '{}' should not be specified in template.".format(param))
                    if param in unused_params:
                        unused_params.remove(param)
                    if param in spec:
                        template = template.replace(replacement, map_replacement(spec, param))
                    else:
                        empties.add(replacement)
                if empties:
                    raise InvalidYAMLException("params {} are not specified in 'default' or 'choice'".format(empties))

                # prepare suffix for logging
                if args.no_subdir:
                    log = "log.{}.{}.{}".format(command, spec["_time"], spec["_name"])
                else:
                    log = "log.{}.{}".format(command, spec["_time"])
                log = shell_arg(os.path.join(spec["_output"], log))

                if args.debug:
                    # log_path may contain tokens that should be escaped in shell.
                    # os.makedirs implicitly handle it, here we should handle it explicitly.
                    suffix = "2>&1 | tee {}".format(log) + "; exit ${PIPESTATUS[0]}"
                else:
                    suffix = "> {} 2>&1".format(log)
                scripts[command] = template + " " + suffix
            orphans.update(unused_params)
            tasks.append({"spec": spec, "scripts": scripts})
    color_print("Orphan params: {}".format(orphans), "red")
    if args.debug:
        tasks = tasks[:1]
        args.force = True
    return tasks


async def build_worker(tasks, queue, resource, skips, fails, force=False, dry_run=False):
    while True:
        index = await queue.get()
        spec, scripts = tasks[index]["spec"], tasks[index]["scripts"]
        prefix = "CUDA_VISIBLE_DEVICES={}".format(resource)
        for command, script in scripts.items():
            scripts[command] = " ".join([prefix, script])
        spec["_scripts"] = scripts

        os.makedirs(spec["_output"], exist_ok=True)
        # dump param
        with edit_yaml(spec["_output"], "param") as prev_spec:
            if not prev_spec:
                prev_spec.update(spec)
            else:
                for key, value in spec.items():
                    # don't overwrite old commands
                    if key == "_scripts":
                        prev_spec[key].update(spec[key])
                    else:
                        prev_spec[key] = spec[key]
        for command, script in scripts.items():
            info = "{:8}:{:2d}/{:2d}, {}".format(command, index + 1, queue.maxsize, shell_arg(spec["_output"]))
            try:
                with edit_yaml(spec["_output"], "stat") as stat:
                    if command in stat and stat[command] in ["finished", "running"] and not force:
                        color_print("SKIP " + info, "green")
                        skips.append(spec["_output"])
                        continue
                    else:
                        stat[command] = "running"
                info = "gpu: {}, ".format(resource) + info
                print("START   " + info)
                if dry_run:
                    print(script)
                    await asyncio.sleep(0.05)
                else:
                    process = await asyncio.create_subprocess_shell(script, executable='/bin/bash')
                    await process.wait()
                    code = process.returncode
                    stat_path = os.path.join(spec["_output"], "stat")
                    with edit_yaml(spec["_output"], "stat") as stat:
                        if code != 0:
                            stat[command] = "failed"
                            color_print("FAIL    " + info, "red")
                            fails.append(spec["_output"])
                            break
                        else:
                            stat[command] = "finished"
            except Exception as exception:
                with edit_yaml(spec["_output"], "stat") as stat:
                    stat[command] = "failed"
                    fails.append(spec["_output"])
                raise exception
            except asyncio.CancelledError as error:
                with edit_yaml(spec["_output"], "stat") as stat:
                    stat[command] = "failed"
                    fails.append(spec["_output"])
                raise error
        queue.task_done()


async def run_all(args, tasks, resources):
    # populate the task queue
    task_num = len(tasks)
    cmd_num = sum(len(task["scripts"]) for task in tasks)
    print("Tasks: {}, Commands: {}".format(task_num, cmd_num))
    queue = asyncio.Queue(maxsize=task_num)
    for index in range(task_num):
        queue.put_nowait(index)
    # build workers that consuming tasks in task_
    skips = []
    fails = []
    workers = []
    loop = asyncio.get_event_loop()
    for resource in resources:
        workers.append(loop.create_task(build_worker(tasks, queue, resource, skips, fails, force=args.force,
                                                     dry_run=args.dry_run)))
    await queue.join()

    for worker in workers:
        worker.cancel()
    await asyncio.gather(*workers, return_exceptions=True)
    if skips:
        color_print("Skip tasks: {}/{}".format(len(skips), len(tasks)), "green")
        for name in skips:
            color_print('    {}'.format(name), "green")

    if fails:
        color_print("Failed tasks: {}/{}".format(len(skips), len(tasks)), "red")
        for name in fails:
            color_print('    {}'.format(name), "red")
    else:
        color_print("No task failed.", "green")


def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-o", "--output", default="output", help="output directory of all experiments")
    parser.add_argument("-y", "--yaml", default="params.yaml",
                        help="yaml configuration file")
    parser.add_argument("-t", "--title", default=None, help="choose param choices with specified title to sweep")
    parser.add_argument("-d", "--debug", default=False, action='store_true',
                        help="debug mode: only run the first task, log will be directed to stdout.")
    parser.add_argument("--dry-run", default=False, action='store_true',
                        help="dry run mode: only print the scripts to be run.")
    parser.add_argument("-c", "--command", default=None, type=str, nargs="+",
                        help="choose which command to run, by default run all commands")
    parser.add_argument("-f", "--force", default=False, action="store_true",
                        help="whether to overwrite tasks successfully ran")
    parser.add_argument("-r", "--resource", default="", nargs="+",
                        help="override resources in params.yaml with a space separate list, "
                             "for example `-r 1,2 3,4` gives ['1,2', '3,4']")
    parser.add_argument("--sample", default=None, type=int,
                        help="number of random samples from each param choice, by default all params choices are ran")
    parser.add_argument("--no-subdir", default=False, action="store_true",
                        help="do not create separated directory for each param choice")

    args = parser.parse_args()
    resources, templates, aliases, defaults, choices = load_yaml(args)
    tasks = build_tasks(args, templates, aliases, defaults, choices)

    os.makedirs(args.output, exist_ok=True)
    yaml_bak_path = os.path.join(args.output, args.yaml if args.title is None else args.yaml + "." + args.title)
    shutil.copyfile(args.yaml, yaml_bak_path)
    run(run_all(args, tasks, resources))



import yaml
from .misc import color_print


class InvalidYAMLException(Exception):
    pass


def parse_config(config):
    def safe_load(key, dtype, required=False):
        if key not in config and required:
            raise InvalidYAMLException("'{}' should be in the first doc.".format(key))

        value = config.get(key, dtype())
        if not isinstance(value, dtype):
            raise InvalidYAMLException("'{}' should be a {}.".format(key, dtype))

        return value

    templates = safe_load("template", dict, required=True)
    if not templates:
        raise InvalidYAMLException("No template exists.")
    resources = safe_load("resource", list, required=True)

    def not_emtpy_dict(name, obj):
        if not isinstance(obj, dict):
            raise InvalidYAMLException("{} should be a dict, not {}.".format(name, type(obj)))
        else:
            if not obj:
                raise InvalidYAMLException("{} should be a nonempty dict.".format(name))

    # alias:
    #   alias1:
    #     case1: { a: 1, b: 1 }
    #     case2: { a: 2, b: 2 }
    aliases = safe_load("alias", dict)
    for key, value in aliases.items():
        not_emtpy_dict("alias[{}]".format(key), value)
        ks = set()  # params to be filled
        for k, v in value.items():
            not_emtpy_dict("alias[{}][{}]".format(key, k), v)
            current_ks = set(v.keys())
            if not ks:
                ks.update(current_ks)
            elif ks != current_ks:
                raise InvalidYAMLException("All entries in alias[{}][{}] should have the same target keys, "
                                           "found difference: {} and {}".format(key, k, ks, current_ks))

    defaults = safe_load("default", dict)
    return templates, resources, aliases, defaults


def filter_choices(title, choices):
    if not choices:
        raise InvalidYAMLException("No choices available.")
    if title is not None:
        title_unset = True  #
        title_not_exist = True
        temp_choices = []
        for choice in choices:
            if "_title" in choice:
                title_unset = False
                if choice["_title"] == title:
                    title_not_exist = False
                    temp_choices.append(choice)
        choices = temp_choices
        if title_unset:
            raise InvalidYAMLException("'_title' not specified in any choices")
        if title_not_exist:
            raise InvalidYAMLException("'_title'={} not found in any choices".format(title))
        color_print("Run choices with title '{}'".format(title), "green")
        color_print(yaml.dump_all(choices, default_flow_style=True, explicit_start=True), "green")
    for choice in choices:
        if "_title" in choice:
            del choice["_title"]
    return choices


def load_yaml(args):
    # parse yaml file
    with open(args.yaml, "r") as fin:
        docs = list(yaml.load_all(fin, Loader=yaml.FullLoader))
    if not docs:
        raise InvalidYAMLException("empty yaml file.")

    templates, resources, aliases, defaults = parse_config(docs[0])

    if args.resource:
        print("Override resource to {}".format(repr(args.resource)))
        resources = args.resource
    resources = [str(i) for i in resources]
    # we dub each doc specifying a grid sweep of different params a "choice"
    choices = filter_choices(args.title, docs[1:])
    return resources, templates, aliases, defaults, choices



import setuptools

setuptools.setup(
        name='mlrunner',
        version='0.5.5',
        author='Tony Ou',
        author_email='simtony2@gmail.com',
        url='https://github.com/simtony/runner',
        description="A light-weight script for gpu experiments.",
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: POSIX :: Linux",
        ],
        install_requires=['pyyaml', 'tabulate'],
        packages=setuptools.find_packages(),
        python_requires='>=3.7',
        entry_points={
            'console_scripts': [
                'run = runner.run:main'
            ]
        }
)
